import itertools

from django.core import validators
from django.db import models
from django.utils.functional import cached_property
from prices import Money, TaxedMoney

from . import forms
from .validators import MoneyPrecisionValidator


class MoneyField(models.DecimalField):

    description = 'A field that stores an amount of money'

    def __init__(self, verbose_name=None, currency=None, **kwargs):
        self.currency = currency
        super(MoneyField, self).__init__(verbose_name, **kwargs)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if isinstance(value, Money):
            if value.currency != self.currency:
                raise ValueError('Invalid currency: %r (expected %r)' % (
                    value.currency, self.currency))
            return value
        value = super(MoneyField, self).to_python(value)
        if value is None:
            return value
        return Money(value, self.currency)

    def get_prep_value(self, value):
        value = self.to_python(value)
        if value is not None:
            return value.amount
        return value

    def get_db_prep_save(self, value, connection):
        value = self.get_prep_value(value)
        db_value = connection.ops.adapt_decimalfield_value
        return db_value(value, self.max_digits, self.decimal_places)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        if value is not None:
            return value
        return super(MoneyField, self).value_to_string(value)

    def formfield(self, **kwargs):
        defaults = {'currency': self.currency, 'form_class': forms.MoneyField}
        defaults.update(kwargs)
        return super(MoneyField, self).formfield(**defaults)

    def get_default(self):
        default = super(MoneyField, self).get_default()
        return self.to_python(default)

    @cached_property
    def validators(self):
        validators = list(
            itertools.chain(self.default_validators, self._validators))
        return validators + [MoneyPrecisionValidator(
            self.currency, self.max_digits, self.decimal_places)]

    def deconstruct(self):
        name, path, args, kwargs = super(MoneyField, self).deconstruct()
        kwargs['currency'] = self.currency
        return name, path, args, kwargs


class TaxedMoneyField(object):

    description = (
        'A field that combines net and gross fields values into TaxedMoney.')
    empty_values = list(validators.EMPTY_VALUES)

    # Field flags
    auto_created = False
    blank = True
    concrete = False
    editable = False

    is_relation = False
    remote_field = None

    def __init__(
            self, net_field='price_net', gross_field='price_gross',
            verbose_name=None, **kwargs):
        self.net_field = net_field
        self.gross_field = gross_field

        self.column = None
        self.primary_key = False

    def __str__(self):
        return ('TaxedMoneyField(net_field=%s, gross_field=%s)' % (
            self.net_field, self.gross_field))

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        net_val = getattr(instance, self.net_field)
        gross_val = getattr(instance, self.gross_field)
        return TaxedMoney(net_val, gross_val)

    def __set__(self, instance, value):
        net = None
        gross = None
        if value is not None:
            net = value.net
            gross = value.gross
        setattr(instance, self.net_field, net)
        setattr(instance, self.gross_field, gross)

    def contribute_to_class(self, cls, name, **kwargs):
        self.attname = self.name = name
        self.model = cls
        cls._meta.add_field(self, private=True)
        setattr(cls, name, self)

    def clean(self, value, model_instance):
        # Shortcircut clean() because Django calls it on all fields with
        # is_relation = False, but we rely on our net and gross fields for
        # actual validation.
        return value



import itertools

from django import forms
from prices import Money

from .validators import (
    MaxMoneyValidator, MinMoneyValidator, MoneyPrecisionValidator)
from .widgets import MoneyInput

__all__ = ('MoneyField', 'MoneyInput')


class MoneyField(forms.DecimalField):
    def __init__(
            self, currency, widget=MoneyInput, max_value=None, min_value=None,
            max_digits=None, decimal_places=None, validators=(),
            *args, **kwargs):
        self.currency = currency
        if isinstance(widget, type):
            widget = widget(currency=self.currency, attrs={
                'type': 'number', 'step': 'any'})
        super(MoneyField, self).__init__(*args, widget=widget, **kwargs)

        self.validators = list(itertools.chain(
            self.default_validators, validators))
        self.validators.append(MoneyPrecisionValidator(
            currency, max_digits, decimal_places))
        if max_value is not None:
            self.validators.append(MaxMoneyValidator(max_value))
        if min_value is not None:
            self.validators.append(MinMoneyValidator(min_value))

    def to_python(self, value):
        value = super(MoneyField, self).to_python(value)
        if value is None:
            return value
        return Money(value, self.currency)

    def validate(self, value):
        if value is None:
            super(MoneyField, self).validate(value)
        else:
            if not isinstance(value, Money):
                raise Exception('%r is not a valid price' % (value,))
            if value.currency != self.currency:
                raise forms.ValidationError(
                    'Invalid currency: %r (expected %r)' % (
                        value.currency, self.currency))
            super(MoneyField, self).validate(value.amount)

    def has_changed(self, initial, data):
        if not isinstance(initial, Money):
            initial = self.to_python(initial)
        return super(MoneyField, self).has_changed(initial, data)



from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def amount(obj):
    result = '%s <span class="currency">%s</span>' % (obj.amount, obj.currency)
    return mark_safe(result)


@register.filter
def discount_amount_for(discount, price):
    return discount(price) - price



from django import forms
from django.template.loader import render_to_string
from prices import Money

__all__ = ['MoneyInput']


class MoneyInput(forms.TextInput):
    template = 'prices/widget.html'
    input_type = 'number'

    def __init__(self, currency, *args, **kwargs):
        self.currency = currency
        super(MoneyInput, self).__init__(*args, **kwargs)

    def format_value(self, value):
        if isinstance(value, Money):
            return value.amount
        return value

    def render(self, name, value, attrs=None):
        widget = super(MoneyInput, self).render(name, value, attrs=attrs)
        return render_to_string(self.template, {
            'widget': widget, 'value': value, 'currency': self.currency})



import re
from decimal import Decimal, InvalidOperation

from babel.core import Locale, UnknownLocaleError, get_global
from babel.numbers import format_currency
from django import template
from django.conf import settings
from django.utils.safestring import mark_safe
from django.utils.translation import get_language, to_locale

from django_babel.templatetags.babel import currencyfmt

register = template.Library()


def get_currency_fraction(currency):
    fractions = get_global('currency_fractions')
    try:
        fraction = fractions[currency]
    except KeyError:
        fraction = fractions['DEFAULT']
    return fraction[0]


def format_price(value, currency, html=False):
    """
    Format decimal value as currency
    """
    try:
        value = Decimal(value)
    except (TypeError, InvalidOperation):
        return ''

    locale, locale_code = get_locale_data()
    pattern = locale.currency_formats.get('standard').pattern

    if html:
        pattern = re.sub(
            '(\xa4+)', '<span class="currency">\\1</span>', pattern)

    result = format_currency(
        value, currency, format=pattern, locale=locale_code)
    return mark_safe(result)


def get_locale_data():
    language = get_language()
    if not language:
        language = settings.LANGUAGE_CODE
    locale_code = to_locale(language)
    locale = None
    try:
        locale = Locale.parse(locale_code)
    except (ValueError, UnknownLocaleError):
        # Invalid format or unknown locale
        # Fallback to the default language
        language = settings.LANGUAGE_CODE
        locale_code = to_locale(language)
        locale = Locale.parse(locale_code)
    return locale, locale_code


@register.filter
def amount(obj, format='text'):
    if format == 'text':
        return format_price(
            obj.amount, obj.currency, html=False)
    if format == 'html':
        return format_price(
            obj.amount, obj.currency, html=True)
    return currencyfmt(obj.amount, obj.currency)



from babel.numbers import get_currency_precision, is_currency

from django.core.validators import (
    DecimalValidator, MaxValueValidator, MinValueValidator, ValidationError)

from .templatetags.prices_i18n import format_price


class MoneyPrecisionValidator(DecimalValidator):
    def __init__(self, currency, *args):
        self.currency = currency
        super(MoneyPrecisionValidator, self).__init__(*args)

    def __call__(self, other):
        if self.currency != other.currency:
            raise ValueError('Invalid currency: %r (expected %r)' % (
                other.currency, self.currency))

        value = other.amount
        super(MoneyPrecisionValidator, self).__call__(value)

        if is_currency(self.currency):
            currency_precision = get_currency_precision(self.currency)
            exponent = value.as_tuple()[-1]
            if exponent >= 0:
                decimals = 0
            else:
                decimals = abs(exponent)
            if decimals > currency_precision:
                raise ValidationError(
                    self.messages['max_decimal_places'],
                    code='max_decimal_places',
                    params={'max': currency_precision})


class MoneyValueValidator:
    def __call__(self, value):
        cleaned = self.clean(value)
        if self.compare(cleaned, self.limit_value):
            currency = self.limit_value.currency
            params = {
                'limit_value': format_price(self.limit_value.amount, currency),
                'show_value': format_price(cleaned.amount, currency),
                'value': format_price(value.amount, currency)}
            raise ValidationError(self.message, code=self.code, params=params)


class MaxMoneyValidator(MoneyValueValidator, MaxValueValidator):
    pass


class MinMoneyValidator(MoneyValueValidator, MinValueValidator):
    pass



#! /usr/bin/env python
from setuptools import setup

CLASSIFIERS = [
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Software Development :: Libraries :: Python Modules']


setup(
    name='django-prices',
    author='Mirumee Software',
    author_email='hello@mirumee.com',
    description='Django fields for the prices module',
    license='BSD',
    version='1.0.0',
    url='https://github.com/mirumee/django-prices',
    packages=['django_prices', 'django_prices.templatetags'],
    include_package_data=True,
    classifiers=CLASSIFIERS,
    install_requires=[
        'Babel>=2.2', 'Django<2.1.0', 'django-babel', 'prices>=1.0.0'],
    platforms=['any'],
    zip_safe=False)
